from openpack.basepack import DefaultNamed, Part


class HeaderPart(DefaultNamed, Part):
    content_type = (
        "application/" "vnd.openxmlformats-officedocument.wordprocessingml.header+xml"
    )
    rel_type = (
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/header"
    )
    default_name = '/word/header.xml'


class FooterPart(DefaultNamed, Part):
    content_type = (
        "application/" "vnd.openxmlformats-officedocument.wordprocessingml.footer+xml"
    )
    rel_type = (
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/footer"
    )
    default_name = '/word/footer.xml'



import six

from paradocx.util import w
from paradocx.package import WordPackage


class Document(WordPackage):
    def paragraph(self, text=None, style=None):
        p = paragraph(text, style=style)
        self.start_part.append(p)
        return p

    def table(self, data=None, style=None):
        tbl = table(data, style=style)
        self.start_part.append(tbl)
        return tbl

    @property
    def paragraphs(self):
        return self.start_part.body.findall(w['p'])


def run(text=None, bold=False, italic=False, font=None):
    rPr = w.rPr()
    if bold:
        rPr.append(w.b())
    if italic:
        rPr.append(w.i())
    if font:
        rFont = w.rFont()
        rFont.attrib[w['ascii']] = font
        rPr.append(rFont)
    r = w.r()
    if len(rPr):
        r.append(rPr)
    if text:
        r.append(w.t(six.text_type(text)))
    return r


def paragraph(text=None, style=None, pagebreak=None):
    p = w.p()
    subs = []
    pPr = w.pPr()
    if style:
        s = w.pStyle()
        s.attrib[w['val']] = style
        pPr.append(s)
    if pagebreak:
        pPr.append(w.sectPr())
    if len(pPr):
        subs.append(pPr)
    if text:
        if isinstance(text, six.string_types):
            text = six.text_type(text)
            subs.append(w.r(w.t(text)))
        elif hasattr(text, 'tag'):
            subs.append(text)
    p.extend(subs)
    return p


def table(data=None, style=None):
    tbl = w.tbl()
    tblPr = w.tblPr()
    tbl.append(tblPr)
    data = data or []
    for cells in data:
        tbl.append(w.tr(*[w.tc(paragraph(value)) for value in cells]))
    if style:
        s = w.tblStyle()
        s.attrib[w['val']] = style
        tblPr.append(s)
    return tbl



"""
Tests for paradocx.image
"""

from __future__ import with_statement

import mimetypes
import warnings
import pytest

from paradocx.image import ImagePart


def test_content_types():
    part = ImagePart(None, '/word/media/foo.jpg')
    jpeg_types = ('image/jpeg', 'image/pjpeg')
    assert part.content_type in jpeg_types
    part = ImagePart(None, '/word/media/bar.jpeg')
    assert part.content_type in jpeg_types
    with pytest.warns(UserWarning):
        part = ImagePart(None, '/foo/bar/strange eps file.eps')
    assert part.content_type == 'application/postscript'
    part = ImagePart(None, '/name without extension')
    assert part.content_type is None
    part = ImagePart(None)
    assert part.content_type == mimetypes.guess_type(part.name)[0]


def test_set_content_type():
    # if the content_type is set, setting .name shouldn't reset it
    part = ImagePart(None)
    part.content_type = 'image/png'
    part.name = '/foo.jpeg'
    assert part.content_type == 'image/png'


def test_interoperability_warning():
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger interoperability warning
        ImagePart(None, '/foo/bar/some_unsupported_format.gif')
        # Verify the warning
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        msg = str(w[-1].message)
        assert "image/gif" in msg and "not" in msg and "interoperable" in msg



import os
import tempfile

import pytest
import pkg_resources


def get_resource_filename(name):
    return pkg_resources.resource_filename(__name__, os.path.join('resources', name))


def get_resource_stream(name):
    return pkg_resources.resource_stream(__name__, os.path.join('resources', name))


_sample_filename = 'sample.docx'


@pytest.fixture
def sample_stream():
    return get_resource_stream(_sample_filename)


@pytest.fixture
def sample_filename():
    return get_resource_filename(_sample_filename)


@pytest.fixture
def table_data():
    data = [
        ['Name', 'Age'],
        ['Christian', 29],
        ['Sarah', 28],
        ['Curtis', 4],
        ['Eli', 2],
    ]
    return data


@pytest.fixture
def writable_filename(request):
    """
    Whenever a function needs a 'writable_filename', create one, but
    be sure it's cleaned up afterward.
    """
    fobj, name = tempfile.mkstemp()
    os.close(fobj)
    os.remove(name)

    def remove_if_exists():
        if os.path.exists(name):
            os.remove(name)

    request.addfinalizer(remove_if_exists)
    return name



from lxml import etree

from openpack.basepack import DefaultNamed, Part
from .util import docx_namespaces

style_ns = docx_namespaces['w']
STYLES = '{%s}' % style_ns


class StylesPart(DefaultNamed, Part):
    content_type = (
        "application/" "vnd.openxmlformats-officedocument.wordprocessingml.styles+xml"
    )
    rel_type = (
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/styles"
    )
    default_name = '/word/styles.xml'

    def load(self, xml):
        lookup = etree.ElementNamespaceClassLookup()
        parser = etree.XMLParser()
        parser.set_element_class_lookup(lookup)
        namespace = lookup.get_namespace(style_ns)
        namespace.update(dict(style=Style))
        self._data = etree.fromstring(xml, parser=parser)

    data = property(lambda self: self._data, load)

    def get_styles(self):
        return self._data.findall(STYLES + 'style')

    def replace_style_id(self, orig, repl):
        # replace all instances of a style id with another
        for style in self.get_styles():
            if style.id == orig:
                style.id = repl
            if style.based_on_id == orig:
                style.based_on_id = repl
            if style.next_id == orig:
                style.next_id = repl

    def replace_styles(self, other):
        """
        Replace all of the styles in self with all of the styles in
        the other StylesPart, but reuse style IDs from the self for any
        styles with the same name.
        """
        self_style_ids = dict((style.name, style.id) for style in self.get_styles())
        for style in other.get_styles():
            if style.name in self_style_ids:
                orig_style_id = style.id
                repl_style_id = self_style_ids[style.name]
                other.replace_style_id(orig_style_id, repl_style_id)
        self._data = other._data
        # todo: make sure there are no duplicate IDs

    def get_style_by_id(self, id):
        styles = dict((style.id, style) for style in self.get_styles())
        return styles[id]


class Style(etree.ElementBase):
    def _get_id(self):
        return self.attrib[STYLES + 'styleId']

    def _set_id(self, id):
        self.attrib[STYLES + 'styleId'] = id

    id = property(_get_id, _set_id)

    @property
    def name(self):
        return self.find(STYLES + 'name').attrib[STYLES + 'val']

    def _get_based_on_id(self):
        node = self.find(STYLES + 'basedOn')
        if node is not None:
            return node.attrib[STYLES + 'val']

    def _set_based_on_id(self, id):
        self.find(STYLES + 'basedOn').attrib[STYLES + 'val'] = id

    based_on_id = property(_get_based_on_id, _set_based_on_id)

    def _get_next_id(self):
        node = self.find(STYLES + 'next')
        if node is not None:
            return node.attrib[STYLES + 'val']

    def _set_next_id(self, id):
        self.find(STYLES + 'next').attrib[STYLES + 'val'] = id

    next_id = property(_get_next_id, _set_next_id)



"""Test the sample.docx provided by ooxmldeveloper.org."""

import os
import posixpath
import random
import hashlib

from paradocx.package import WordPackage
from paradocx.document import DocumentPart
from paradocx.styles import StylesPart, Style
from paradocx.util import dcterms


def test_open(sample_stream):
    pkg = WordPackage.from_stream(sample_stream)
    assert not hasattr(pkg, 'filename')


def test_open_file(sample_filename):
    pkg = WordPackage.from_file(sample_filename)
    assert pkg.filename == sample_filename


def test_start_part(sample_stream):
    pkg = WordPackage.from_stream(sample_stream)
    assert isinstance(pkg.start_part, DocumentPart)
    assert pkg.start_part.name == '/word/document.xml'
    assert pkg.start_part.name in pkg


def test_create_empty_styles():
    StylesPart(None)


def test_styles(sample_stream):
    pkg = WordPackage.from_stream(sample_stream)
    styles_path = posixpath.join('/word', 'styles.xml')
    assert styles_path in pkg.keys()
    srel = pkg.start_part.related(StylesPart.rel_type)[0]
    assert srel == pkg[styles_path]


def test_core_props(sample_stream):
    pkg = WordPackage.from_stream(sample_stream)
    assert pkg.core_properties
    assert id(pkg.core_properties.element.find(dcterms['modified']))


def test_style_class(sample_stream):
    pkg = WordPackage.from_stream(sample_stream)
    styles_part = pkg['/word/styles.xml']
    styles = styles_part.get_styles()
    assert len(styles)
    assert all(isinstance(style, Style) for style in styles)
    assert len(styles) == len(set(style.id for style in styles))
    style_names = [style.name for style in styles]
    assert len(set(style_names)) == len(styles)


def test_replace_style_id(sample_stream):
    pkg = WordPackage.from_stream(sample_stream)
    styles_part = pkg['/word/styles.xml']
    styles_part.replace_style_id('DefaultParagraphFont', 'NewId')
    # header char is based on the default paragraph font
    header_char = styles_part.get_style_by_id('HeaderChar')
    assert header_char.based_on_id == 'NewId'


def test_replace_styles_by_name(sample_stream):
    pkg = WordPackage.from_stream(sample_stream)
    orig_styles_part = pkg['/word/styles.xml']
    new_styles_part = StylesPart(None, '/who/cares', data=orig_styles_part.dump())
    # change an id in the "new" styles
    style = random.choice(new_styles_part.get_styles())
    random_bytes = os.urandom(100)
    new_id = hashlib.md5(random_bytes).hexdigest()
    orig_style_id, style.id = style.id, new_id
    style_name = style.name
    # tag the element, so we're sure it gets copied to the orig styles
    style.attrib['tag'] = 'test-replace-styles'
    orig_styles_part.replace_styles(new_styles_part)
    assert new_id not in (style.id for style in orig_styles_part.get_styles())
    assert style_name in (style.name for style in orig_styles_part.get_styles())
    style = orig_styles_part.get_style_by_id(orig_style_id)
    assert b'test-replace-styles' in orig_styles_part.dump()



from lxml.etree import fromstring
from openpack.basepack import Part
from openpack.officepack import OfficePackage

from .util import w, expand_namespace as EN


class DocumentPart(Part):
    content_type = (
        "application/"
        "vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml"
    )
    rel_type = OfficePackage.main_rel

    def __init__(self, *args, **kwargs):
        Part.__init__(self, *args, **kwargs)
        self.data = w.document()
        self.body = w.body()
        self.data.append(self.body)
        if 'data' in kwargs:
            self.load(kwargs['data'])

    def load(self, xml):
        self.data = fromstring(xml)
        self.body = self.data.find(EN('w:body'))

    def append(self, xml_element):
        self.body.append(xml_element)



"""Test the sample.docx provided by ooxmldeveloper.org."""

import io

from paradocx import Document


def test_add_paragraph():
    doc = Document()
    assert not doc.paragraphs
    doc.paragraph('Hello World!')
    assert doc.paragraphs


def test_add_table(table_data):
    doc = Document()
    doc.table(table_data)


def test_save():
    doc = Document()
    stream = io.BytesIO()
    doc._store(stream)
    assert stream.tell() > 0


def test_read_paragraph(writable_filename):
    doc = Document()
    doc.paragraph('Hello World!')
    doc.save(writable_filename)
    doc = Document.from_file(writable_filename)
    assert b'Hello World' in doc.start_part.dump()


def test_read_table(table_data, writable_filename):
    doc = Document()
    doc.table(table_data)
    doc.save(writable_filename)
    doc = Document.from_file(writable_filename)
    assert b'Christian' in doc.start_part.dump()



from openpack.basepack import DefaultNamed, Part
import mimetypes
import warnings


class ImagePart(DefaultNamed, Part):
    rel_type = (
        'http://schemas.openxmlformats.org/officeDocument/2006/relationships/image'
    )
    default_name = '/word/media/image.jpeg'

    """
    ECMA-376 3nd Edition Part 1 Page 170 states:
    A producer that wants interoperability should use
    one of the following standard formats:
        - image/png ISO/IEC 15948:2003, http://www.libpng.org/pub/png/spec/
        - image/jpeg, http://www.w3.org/Graphics/JPEG
    """
    interoperability_types = ['image/png', 'image/jpeg']

    def _set_name(self, name):
        super(ImagePart, self)._set_name(name)
        self._guess_mime_type(name)

    name = property(Part._get_name, _set_name)

    def _guess_mime_type(self, name):
        """
        When setting the name, guess the mime type from the extension.
        Set the content_type for this instance only if a content_type is not
        already defined (this allows an instance to have a content-type pre-
        defined or for a subclass to define the content type, and it will not
        be overridden by the guessed type).
        """
        ct, _ = mimetypes.guess_type(name)
        if ct and not self.content_type:
            self.content_type = ct
            if ct not in self.interoperability_types:
                warnings.warn(
                    "Image type %s is not guaranteed to be interoperable" % ct
                )



from openpack.basepack import DefaultNamed, Part


class NumberingPart(DefaultNamed, Part):
    content_type = (
        'application/'
        'vnd.openxmlformats-officedocument.wordprocessingml.numbering+xml'
    )
    rel_type = (
        'http://schemas.openxmlformats.org'
        '/officeDocument/2006/relationships/numbering'
    )
    default_name = '/word/numbering.xml'


class AbstractNumbering(object):
    def __init__(self, id, **props):
        self.id = id
        self.properties = props


class Numbering(object):
    pass



import importlib

from openpack.basepack import ContentType
from openpack.officepack import OfficePackage

from . import document


def initialize():
    """
    Import various modules so their Part classes are present.
    """
    for mod in 'document numbering styles headerfooter'.split():
        importlib.import_module('paradocx.' + mod)


initialize()


class WordPackage(OfficePackage):
    def __init__(self):
        """
        Construct a WordPackage.
        """
        super(WordPackage, self).__init__()
        self.quickstart()

    def _load(self, *args, **kwargs):
        # re-initialize this package without the quickstart
        super(WordPackage, self).__init__()
        super(WordPackage, self)._load(*args, **kwargs)

    def quickstart(self):
        self.content_types.add(
            ContentType.Default(
                "application/xml",
                "xml",
            )
        )
        self.content_types.add(
            ContentType.Default(
                "application/vnd.openxmlformats-package.relationships+xml",
                "rels",
            )
        )
        start = document.DocumentPart(self, '/word/document.xml')
        self.content_types.add(
            ContentType.Override(
                start.content_type,
                start.name,
            )
        )
        self[start.name] = start
        self.relate(start)


if __name__ == '__main__':
    import sys
    import codecs
    from textwrap import dedent
    from this import s as zen_encoded
    from openpack.basepack import CoreProperties

    trun = dedent(
        """
        <w:p>
            <w:r>
                <w:t>%s</w:t>
            </w:r>
        </w:p>
        """
    ).lstrip()
    raw_zen = codecs.decode(zen_encoded, 'rot13')
    zen = "".join(trun % line for line in raw_zen.splitlines())

    body = (
        dedent(
            """
        <?xml version="1.0" encoding="utf-8"?>
        <w:document
        xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
            <w:body>
                <w:p>
                    <w:r><w:t>Generated by:</w:t></w:r>
                </w:p>
                <w:p>
                    <w:r><w:t>Python %s</w:t></w:r>
                </w:p>
                %s
            </w:body>
        </w:document>
        """
        ).lstrip()
        % (sys.version, zen)
    )
    op = WordPackage()
    op.start_part.data = body
    cp = CoreProperties(op, '/docProps/core.xml')
    cp.encoding = 'utf-8'
    op.add(cp)
    op.relate(cp)
    cp.creator = "Testy McTest"
    op.save('test.docx')



from lxml.etree import fromstring, tostring
from lxml import builder

from openpack.basepack import ooxml_namespaces

docx_namespaces = {
    'w': "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
    'r': 'http://schemas.openxmlformats.org/officeDocument/2006/relationships',
    've': 'http://schemas.openxmlformats.org/markup-compatibility/2006',
    'wp': ('http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing'),
    'v': 'urn:schemas-microsoft-com:vml',
}
docx_namespaces.update(ooxml_namespaces)


def expand_namespace(tag):
    """
    >>> expand_namespace('w:document')
    '{http://schemas.openxmlformats.org/wordprocessingml/2006/main}document'
    """
    namespace, sep, tag = tag.rpartition(':')
    fmt = '{%(namespace)s}%(tag)s' if namespace else '%(tag)s'
    namespace = docx_namespaces[namespace]
    return fmt % vars()


class ElementMaker(builder.ElementMaker):
    def __getitem__(self, name):
        return "%s%s" % (self._namespace, name)


w = ElementMaker(namespace=docx_namespaces['w'], nsmap=docx_namespaces)
dcterms = ElementMaker(namespace=docx_namespaces['dcterms'], nsmap=docx_namespaces)


class TemplateSource(object):
    template = None

    def __init__(self, variables, encoding='utf-8'):
        self.variables = variables
        self.xml = self._from_template()
        self.element = self._to_element()

    def _from_template(self):
        """Use self.template and self.variables to generate XML content."""
        raise NotImplementedError

    def _to_element(self):
        return fromstring(self.xml)

    def __str__(self):
        return tostring(self.element, encoding=self.encoding)



from openpack.basepack import DefaultNamed, Part


class SettingsPart(DefaultNamed, Part):
    content_type = (
        'application/' 'vnd.openxmlformats-officedocument.wordprocessingml.settings+xml'
    )
    rel_type = (
        'http://schemas.openxmlformats.org'
        '/officeDocument/2006/relationships/settings'
    )
    default_name = '/word/settings.xml'
